# SparseFA : simulation
# experiments on simulated data

from vbfa import *
import sys
import scipy as SP
from mlib.utils import *
from sparseFA import *
import scipy
import scipy.io
import os
import logging as L
import pylab as PL
import time
import pdb


###Helper classes and functions####

def calc_accuracy(Z,X):
    """calc predictive accuracy
    Z: responsibilties
    X: truth"""
    Nok = (Z==X).sum()
    return double(Nok)/X.size

def calc_logprob(Z,X):
    ptrue = Z[X]+1E-10
    pfalse= (1-Z)[~X]+1E-10
    return mean(log(concatenate((ptrue,pfalse))))
    pass
    

def evaluate_C(C,sim):
    """evaluate incoming results
    C:  from predictor
    sim: truth"""
    auc = ROC.auroc(labels=sim['Ct'],predictions=C)
    acc = calc_accuracy(C>0.5,sim['Ct'])
    lp  = calc_logprob(C,sim['Ct'])
    fon = double(C.sum())/C.size
    return {'auc':auc,'acc':acc,'fon':fon,'lp':lp}

def evaluate_WX(W,C,X,sim,Ip=None):
    """evaluate the predictive accuracy for product pairs"""
    def getPairs(W,X):
        """create pair arrays"""
        W_ = W.reshape([W.shape[0],1,W.shape[1]]).repeat(X.shape[0],axis=1)
        X_ = X.reshape([1,X.shape[0],X.shape[1]]).repeat(W.shape[0],axis=0)
        return [W_,X_]
    Wt = sim['Wt']
    Xt = sim['Xt']
    #pWC = W*C
    if Ip is None:
        Ip = SP.arange(W.E1.shape[1])
    WE1 = W.E1[:,Ip]
    XE1 = X.E1[:,Ip]
    XS2 = S.zeros_like(XE1)
    WS2 = S.zeros_like(WE1)
    WS2 = W.cov.diagonal(axis1=1,axis2=2)[:,Ip]
    XE2 = (X.cov[:,Ip][Ip,:]).diagonal().reshape([1,-1]).repeat(XS2.shape[0],axis=0)


    #get all pairs
    #truth
    [Wt_,Xt_]  = getPairs(Wt,Xt)
    #mean
    [WE1_,XE1_]= getPairs(WE1,XE1)
    #var
    [WS2_,XS2_] = getPairs(WS2,XS2)

    #total mean:
    mean = WE1_*XE1_
    #total variance
    var  = WS2_ * XE1_**2 + XS2_ * WE1_**2 + WS2_*XS2_

    #mean (truth)
    meant= Wt_ *Xt_
    
    #1. mse
    D2  = (mean-meant)**2
    mse  = (D2.mean())
    #2. mean log prob
    lprob  = -0.5*D2/var - 0.5*S.log(2*pi*var)
    mlp    = lprob.mean()
    #3. reconstruction error
    Yr = dot(X.E1[:,Ip],W.E1[:,Ip].T)
    mseY = (((sim['Y']-Yr)**2).mean())
    return {'mse': mse,'mlp':mlp,'mseY':mseY}
    pass

def accuracy_prior(Pi,sim,nRestarts=1):
    """run an experiment of the prior"""
    RB = S.zeros([1,nRestarts])
    R = {'ACC':RB.copy(),'AUC':RB.copy(),'LOGPROB':RB.copy(),'MSE':RB.copy(),'MSEY':RB.copy()}
    for ir in range(nRestarts):
        #draw sample from prior
        Z= random.rand(Pi.shape[0],Pi.shape[1])<Pi
        #evaluate
        ev = evaluate_C(Z*1,sim)
        R['LOGPROB'][0,ir] = calc_logprob(Pi,sim['Ct'])
        R['ACC'][0,ir] = ev['acc']
        R['AUC'][0,ir] = ev['auc']
    R['T'] = SP.array([0])
    return R
    pass


def accuracy_experiment(vbfa,nIblock =2,nRestarts=1,sim=None,init=None):
    """helper function to run an accuracy experiment
    vbfa: vbfa object to use (needs to be initialised)
    Xt: true network for evaluation
    nIblock: number of itertions per block (2)
    nRestarts: number of restarts
    init: init block for model
    """
    nIterations = vbfa.nIterations
    #number of blocks?
    nblocks     = nIterations/nIblock
    RB = S.zeros([nblocks,nRestarts])
    #create result structure
    R = {'T':RB.copy(),'ITER':RB.copy(),'ACC':RB.copy(),'AUC':RB.copy(),'LOGPROB':RB.copy(),'MSE':RB.copy(),'MSEY':RB.copy()}
    for ir in range(nRestarts):
        #measure start-time
        Tstart = time.clock()
        #init ?
        if init is not None:
            vbfa.init(**init)
        for i in xrange(nblocks):
            vbfa.iterate(nIterations=nIblock)
            #get the prdictions and calculate precisions
            #hack:
            if vbfa.permutation_move:
                Ip = vbfa.W.Ilabel.argsort()
            else:
                Ip = SP.arange(vbfa.components)
            C = vbfa.W.C[:,Ip,1]
            # calc pred. accuracy on indictor
            evC = evaluate_C(C,sim)
            # calc pred. accuruacy for pairs of weights, X
            evWX= evaluate_WX(vbfa.W,C,vbfa.S,sim,Ip)
            delta_t = time.clock()-Tstart
            #save results
            R['ACC'][i,ir] = evC['acc']
            R['AUC'][i,ir] = evC['auc']
            R['LOGPROB'][i,ir] = evC['lp']
            R['MSE'][i,ir] = evWX['mse']
            R['MSEY'][i,ir] = evWX['mseY']
            R['T'][i,ir] = delta_t
            R['ITER'][i,ir] = i*nIblock
            print "accuracy: %f (fraction on on %f)" % (evC['acc'],evC['fon'])
    #save global information
    R['name'] = vbfa.getName()
    R['model'] = vbfa
    return R


if __name__ =='__main__':
    L.getLogger().setLevel(L.INFO)
    random.seed(2)

    out_base = './../../out'
    data_base ='./../../Data/JMLR'

    dataset = 'Toy_486_20_20'
#    dataset = 'Toy_800_60_100'

    if 1:
        #load small simulated dataset
        dump_file = os.path.join(data_base,dataset,'dataset.pickle')
        sim = cPickle.load(open(dump_file,'rb'))
        Pi = sim['Pit']
        Y = sim['Y']
        components = Pi.shape[1]

    save = True
    nRestarts   = 3
    permutation_move = False

    ####experiments for small network
    if (len(sys.argv)>1):
        exp_arg = sys.argv[1]
    else:
        exp_arg = ''
    if ((exp_arg=='EP') | (exp_arg=='EPp')):
        sigma_off = [sqrt(0.1),0.1,0.01,0.001,0.0001]
        nIterations = 300
#        sigma_off = [sqrt(0.1),0.1,0.01]
        #nIterations = 50
        nIterations = 50
        exp_name    = 'EP_nonrandomorder'
        randperm    = False
        sparsity = 'EPV'
        #changes for permutation move EP
        if exp_arg=='EPp':
            exp_name = 'EPp_norandomorder'
            permutation_move = True
    elif ((exp_arg=='VB') | (exp_arg=='VBp')):
        exp_name    = 'VB_nonrandomorder'
        sigma_off = [sqrt(0.1),0.1,0.01,0.001,0.0001]
        nIterations = 1000
#        sigma_off = [sqrt(0.1),0.1,0.01]
#        nIterations = 200
        nIterations = 300
        sparsity = 'VB'
        if exp_arg=='VBp':
            exp_name = 'VBp_norandomorder'
            permutation_move = True
    elif exp_arg=='prior':
        exp_name = 'prior'
    else:
        ###debug
        nRestarts = 1
        sigma_off = [1E-3]
        nIterations = 50
        exp_name    = 'EP'
        randperm    = False
        sparsity = 'EPV'

    #do w hav an S2 indicator for batch mode processing ? 
    if(len(sys.argv)>2):
        IS2 = int(sys.argv[2])
    else:
        IS2 = None

    dump_file = os.path.join(out_base,dataset,exp_name+ '.pickle')
    init={'init_data':CGauss(Y),'Pi':Pi}
    initType = 'prior'
    results = []

    if 1:
        priors = {'Eps':{'priors': [1,10]}}
        #0. prior
        if exp_name[0:5]=='prior':
            results = accuracy_prior(Pi,nRestarts=nRestarts,sim=sim)
            results['name'] = 'prior'
            dump_file = os.path.join(out_base,dataset,exp_name)
            dump_file = dump_file + '.pickle' 
            cPickle.dump(results,open(dump_file,'wb'),-1)
        else:
            if IS2 is not None:
                sigma_off = [sigma_off[IS2]]
            for s2 in sigma_off:
                FA = CSparseFA(components=components,sigmaOff=s2,sparsity=sparsity,nIterations=nIterations,permutation_move=permutation_move,priors=priors,initType=initType)
                RFA    = accuracy_experiment(FA,nRestarts=nRestarts,init=init,sim=sim)
                #dump results
                suffix  = '_S2%.1E.pickle' %(s2)
                dump_file1 = os.path.join(out_base,dataset,'full_' + exp_name + suffix)
                dump_file2 = os.path.join(out_base,dataset,'result_' + exp_name + suffix)

                if save:
                    cPickle.dump(RFA,open(dump_file1,'wb'),-1)
                    #now delete object dump
                    del(RFA['model'])
                    cPickle.dump(RFA,open(dump_file2,'wb'),-1)
                

