# Experiments on smith yeast data

from sparseFA import *
from vbfa_fillin import *
import pylab as PL
import scipy
import scipy.io
import os
import logging as L
import pylab as PL

class ExpressionData(object):

    def __init__(self,mat_file,FNR=0,FPR=0):
        #1. load data
        if os.path.exists(mat_file+ '.pickle'):
            data = cPickle.load(open(mat_file+'.pickle','rb'))
        else:
            data = scipy.io.loadmat(mat_file)
            cPickle.dump(data,open(mat_file+'.pickle','wb'),-1)

        self.data = data
        #extract stuff
        self.Y    = data['Y']
        self.Ig   = arange(self.Y.shape[1])
        if(data.has_key('Pi')):
            self.Pi   = data['Pi']
            self.K    = self.Pi.shape[0]
        if(data.has_key('X')):
            self.X   = data['X']
            self.K   = self.X.shape[0]
            #create PI
            self.Pi = S.array(self.X>0,dtype='float')*FPR
            self.Pi[self.X==0] = FNR
            pass
        self.If   = arange(self.Pi.shape[0])
        [self.N,self.D]    = self.Y.shape

    def subset(self,Ngenes=None,NTF=None):
        """create a subset of the expresison matrix, filtering the genes by variance and eliminating TFs without sufficient connectivity"""
        #1. variance
        if Ngenes is not None:
            Ig = argsort(self.Y.std(axis=0)[::-1])
            Ig = Ig[0:Ngenes]
            self.Y = self.Y[:,Ig]
            self.Pi = self.Pi[:,Ig]
            self.X = self.X[:,Ig]
            self.Ig = self.Ig[Ig]
        #2. check tfs
        if NTF is not None:
            If = self.X.sum(axis=1)>NTF
            self.X = self.X[If,:]
            self.Pi = self.Pi[If,:]
            self.K  = self.X.shape[0]
            self.If = self.If[If]
        #3. exlclude genes without any connection
        I0 = self.X.sum(axis=0)==0
        self.Y = self.Y[:,~I0]
        self.X = self.X[:,~I0]
        self.Pi= self.Pi[:,~I0]
        self.Ig = self.Ig[I0]
        
    def normaliseY(self):
        self.Y = self.Y/self.Y.std(axis=0)
        
    def getExpr(self):
        return self.Y

    def getPi(self):
        return self.Pi



class FillInRun(object):
    """a single fill-in run for a given splittng and a set of masks"""

    def __init__(self,data,sr=0.7,fillin_rates=([0.1,0.5,0.8])):
        self.data = data
        perm = random.permutation(data.N)
        Ntrain = int(round(data.N*sr))
        Ntest  = data.N-Ntrain
        self.I_train = perm[0:Ntrain]
        self.I_test  = perm[Ntrain::]
        #create masks
        self.fillin_rates = fillin_rates
        self.masks = []
        for fr in fillin_rates:
            self.masks.append(random.rand(Ntest,self.data.D)>fr)

    def testModels(self,models):
        """test all models in models-array"""

        results = []
        for m in models:
            m_train = m[0]
            m_test  = m[1]
            results.append(self.testModel(m_train,m_test))
        self.results = array(results)
        self.models  = models
        return results

    def testModel(self,model_train,model_test,params_train={},params_test={}):
        """test model model which is assumed to be of a sparseFA type family
        model_train: model to train
        model_test:  model to test with marginal posteriors from model_train as priors
        mask: missing values mask
        params_train: parameters fro training run
        params_test : parameteres for test run
        """
        
        #1. initialise model with trainign data and iterate
        Ytrain = self.data.Y[self.I_train]
        Ytest  = self.data.Y[self.I_test]
        Pi     = self.data.Pi
        model_train.init(CGauss(Ytrain),Pi=Pi.T)
        model_train.iterate(**params_train)
        #2. test model_test for every mask
        rmse = zeros(len(self.masks))
        for i in range(len(self.masks)):
            mask = self.masks[i]            
            model_test.init(CGauss(Ytest),model_train=model_train,mask=mask)
            model_test.iterate(**params_test)
            #3. test
            Yp = model_test.getPrediction().E1
            DY = Ytest-Yp
            rmse[i] = sqrt(((DY[~mask])**2).mean())
        return rmse



def test_diverse1(data,Nf =1):
    """a comprehensive comparison of all sorts of models """
    
    #try that out using a model
    components = data.K
    sigmaOff = 0.1
    nIterations_train = 30
    nIterations_test = 1
    #fill-in model
    model_test      = CVBFA_fillin(nIterations=nIterations_test)

    #mean
    model_mean      = CMean_fillin()
    #VBFA
    model_VB        = CVBFA(components=40,nIterations=nIterations_train)
    model_VB2       = CVBFA(components=40,nIterations=nIterations_train,priors={'Alpha':{'priors':[1E-3,1E1]}})
    model_VB3       = CVBFA(components=40,nIterations=nIterations_train,priors={'Alpha':{'priors':[1E-3,1E1]}}) 
    models_VB       = [[model_VB,model_test],[model_VB2,model_test],[model_VB3,model_test]]

    models_VB       = [[model_VB,model_test]]
    
    #Sparse VB
    model_SVB       = CSparseFA(components=components,sigmaOff=sigmaOff,sparsity='VB',nIterations=nIterations_train)
    #Sparse EP
    model_SEP       = CSparseFA(components=components,sigmaOff=sigmaOff,sparsity='EP',nIterations=nIterations_train)
    #Sparse EPV
    model_SEPV      = CSparseFA(components=components,sigmaOff=sigmaOff,sparsity='EPV',nIterations=nIterations_train)

    sigmaOff=0.01
    #Sparse VB
    model_SVB2       = CSparseFA(components=components,sigmaOff=sigmaOff,sparsity='VB',nIterations=nIterations_train)
    #Sparse EP
    model_SEP2       = CSparseFA(components=components,sigmaOff=sigmaOff,sparsity='EP',nIterations=nIterations_train)
    #Sparse EPV
    model_SEPV2      = CSparseFA(components=components,sigmaOff=sigmaOff,sparsity='EPV',nIterations=nIterations_train)
    
    models_sparse   = [[model_SVB,model_test],[model_SEP,model_test],[model_SEPV,model_test],[model_SVB2,model_test],[model_SEP2,model_test],[model_SEPV2,model_test]]
    

    #array of all models
    models = [[model_mean,model_mean]]
    models.extend(models_VB)
    models.extend(models_sparse)

    #fill-in runs
    fr = zeros(Nf,dtype='object')
    for n in range(Nf):
        fr[n]  = FillInRun(data,fillin_rates=([0.1,0.3,0.5,0.8]))
        fr[n].testModels(models)

    return fr




def test_diverse2(data,Nf =1):
    """a comprehensive comparison of all sorts of models """
    
    #try that out using a model
    components = data.K
    sigmaOffVB = 0.1
    sigmaOffEP = 0.1
    nIterations_train = 70
    nIterations_test = 1


    #fill-in model
    model_test      = CVBFA_fillin(nIterations=nIterations_test)

    #mean
    model_mean      = CMean_fillin()
    #VBFA
    model_VB        = CVBFA(components=components,nIterations=nIterations_train)
    model_VB2       = CVBFA(components=40,nIterations=nIterations_train)
    models_VB       = [[model_VB,model_test],[model_VB2,model_test]]
    models_VB       = [[model_VB2,model_test]]
    
    #Sparse VB
    priorsVB = {'Eps':{'priors': [1,100]}}
    model_SVB       = CSparseFA(components=components,sigmaOff=sigmaOffVB,sparsity='VB',nIterations=nIterations_train,priors=priorsVB)

    #Sparse EPV
    priorsEP = {'Eps':{'priors': [1,10]}}
    model_SEPV      = CSparseFA(components=components,sigmaOff=sigmaOffEP,sparsity='EPV',nIterations=nIterations_train,priors=priorsEP)
    priorsEP = {'Eps':{'priors': [1,100]}}
    model_SEPV2      = CSparseFA(components=components,sigmaOff=sigmaOffEP,sparsity='EPV',nIterations=nIterations_train,priors=priorsEP)

    models_sparse   = [[model_SVB,model_test],[model_SEPV,model_test],[model_SEPV2,model_test]]

    #array of all models
    models = [[model_mean,model_mean]]
    models.extend(models_VB)
    models.extend(models_sparse)
    

    #fill-in runs
    fr = zeros(Nf,dtype='object')
    for n in range(Nf):
        fr[n]  = FillInRun(data,fillin_rates=([0.1,0.3,0.5,0.8]))
        fr[n].testModels(models)
    return fr




def test_diverse3(data,Nf =1):
    """the test that is run in the thesis results"""
    
    #try that out using a model
    components = data.K
    sigmaOffVB = 0.1
    sigmaOffEP = 0.1
    nIterations_train = 80
    nIterations_test = 1


    #fill-in model
    model_test      = CVBFA_fillin(nIterations=nIterations_test)

    #mean
    model_mean      = CMean_fillin()
    #VBFA
    model_VB        = CVBFA(components=components,nIterations=nIterations_train)
    model_VB2       = CVBFA(components=40,nIterations=nIterations_train)
    models_VB       = [[model_VB,model_test],[model_VB2,model_test]]
    models_VB       = [[model_VB2,model_test]]
    
    #Sparse VB
    priorsVB = {'Eps':{'priors': [1,100]}}
    model_SVB       = CSparseFA(components=components,sigmaOff=sigmaOffVB,sparsity='VB',nIterations=nIterations_train,priors=priorsVB)


    #Sparse EPV
    priorsEP = {'Eps':{'priors': [1,10]}}
    model_SEPV      = CSparseFA(components=components,sigmaOff=sigmaOffEP,sparsity='EPV',nIterations=nIterations_train,priors=priorsEP)
    priorsEP = {'Eps':{'priors': [1,100]}}
    model_SEPV2      = CSparseFA(components=components,sigmaOff=sigmaOffEP,sparsity='EPV',nIterations=nIterations_train,priors=priorsEP)

    models_sparse   = [[model_SVB,model_test],[model_SEPV,model_test],[model_SEPV2,model_test]]

    #array of all models
    models = [[model_mean,model_mean]]
    models.extend(models_VB)
    models.extend(models_sparse)
    

    #fill-in runs
    fr = zeros(Nf,dtype='object')
    for n in range(Nf):
        fr[n]  = FillInRun(data,fillin_rates=([0.1,0.3,0.5,0.7,0.9]))
        fr[n].testModels(models)
    return fr



def test_diverse4(data,Nf =1):
    """the test that is run in the thesis results"""
    
    #try that out using a model
    components = data.K
    sigmaOffVB = 0.1
    sigmaOffEP = 0.01
    nIterations_train = 80


    sigmaOffVB = 0.01
    sigmaOffEP = 0.01
    nIterations_train = 40
    nIterations_test = 1


    #fill-in model
    model_test      = CVBFA_fillin(nIterations=nIterations_test)

    #mean
    model_mean      = CMean_fillin()
    #VBFA
    model_VB        = CVBFA(components=components,nIterations=nIterations_train)
    model_VB2       = CVBFA(components=40,nIterations=nIterations_train)
    models_VB       = [[model_VB,model_test],[model_VB2,model_test]]

    models_VB       = [[model_VB2,model_test]]
    
    #Sparse VB
    priorsVB = {'Eps':{'priors': [1,10]}}
    model_SVB       = CSparseFA(components=components,sigmaOff=sigmaOffVB,sparsity='VB',nIterations=nIterations_train,priors=priorsVB)

    #Sparse EPV
    priorsEP = {'Eps':{'priors': [1,10]}}
    model_SEPV      = CSparseFA(components=components,sigmaOff=sigmaOffEP,sparsity='EPV',nIterations=nIterations_train,priors=priorsEP)
    
    models_sparse   = [[model_SVB,model_test],[model_SEPV,model_test]]

    #array of all models
    models = [[model_mean,model_mean]]
    models.extend(models_VB)
    models.extend(models_sparse)
    

    #fill-in runs
    fr = zeros(Nf,dtype='object')
    for n in range(Nf):
        fr[n]  = FillInRun(data,fillin_rates=([0.1,0.3,0.5,0.7,0.9]))
        fr[n].testModels(models)
    return fr



def test_VBEP(data,Nf =1):
    """only apply VB/EP to obtain hidden factors"""
    
    #try that out using a model
    components = data.K
    sigmaOffVB = 0.1
    sigmaOffEP = 0.1
    nIterations_train = 80
    nIterations_test = 1


    #fill-in model
    model_test      = CVBFA_fillin(nIterations=nIterations_test)

    #mean
    model_mean      = CMean_fillin()
    #Sparse EPV
    priorsEP = {'Eps':{'priors': [1,10]}}
    model_SEPV      = CSparseFA(components=components,sigmaOff=sigmaOffEP,sparsity='EPV',nIterations=nIterations_train,priors=priorsEP)


    models_sparse   = [[model_SEPV,model_test]]
    #array of all models
    models = [[model_mean,model_mean]]
    models.extend(models_sparse)
    

    #fill-in runs
    fr = zeros(Nf,dtype='object')
    for n in range(Nf):
        fr[n]  = FillInRun(data,fillin_rates=([0.1,0.3,0.5,0.7,0.9]))
        fr[n].testModels(models)
    return fr



def plotResults(results):
    """plot a results-structure"""
    #1. create large results structure
    [Nmodels,Nfill] = results[0].results.shape
    RA = zeros([Nmodels,Nfill,len(results)])
    for nr in range(len(results)):
        RA[:,:,nr] = results[nr].results
    #2. calculate mean and std
    Rm = RA.mean(axis=2)
    Rs = RA.std(axis=2)
    #3. plot
    X  = results[0].fillin_rates
    plots=[]
    names=[]
    for m in range(Nmodels):
        names.append(results[0].models[m][0].getName())
#        names.append("model: %d" % (m))
        plots.append(PL.errorbar(x=X,y=Rm[m,:],yerr=Rs[m,:],linewidth=2,markersize=12,fmt='-')[0])
    PL.legend(plots,names)
    pass
    

    


if __name__ =='__main__':
    L.getLogger().setLevel(L.DEBUG)

    random.seed(1)

    #data file for small yeast data
    data_small = '../Data/smith08_yeast/small/experiment_data.mat'
    data_large = '../Data/smith08_yeast/large/experiment_data.mat'
    data_gasch = '../Data/gasch_yeast/large/experiment_data.mat'

    data_mnaimneh= './../Data/mnaimneh_yeast/large/experiment_data.mat'

    if 0:
        data = ExpressionData(data_small)
        out_base   = './../out/yeast/small'
        Nf = 1

    Nf = 1
    if 1:
        if 1:
            dump_file ='gasch'
            data = ExpressionData(data_mnaimneh,FNR=0.05,FPR=0.95)
            run = 1
            random.seed(run)
            dump_file ='gasch2_thesis%d' % (run)

            NTF = 20

        if 0:
            dump_file = 'mnaimneh'
            data = ExpressionData(data_mnaimneh,FNR=0.05,FPR=0.95)
            run = 3
            random.seed(run)
            dump_file ='mnaimneh_thesis%d' % (run)

            NTF = 20
        if 0:
            #1. SMITH
            data = ExpressionData(data_large,FNR=0.0048,FPR=0.94)
            data = ExpressionData(data_large,FNR=0.005,FPR=0.95)
            dump_file ='smith_00048_094_stronger2'


            data = ExpressionData(data_large,FNR=0.001,FPR=0.99)

            data = ExpressionData(data_large,FNR=0.05,FPR=0.95)
            run = 3
            random.seed(run)
            dump_file ='smith_thesis_2%d' % (run)
            NTF = 20
        out_base   = './../out/yeast/large'

        data.subset(Ngenes=1500,NTF=NTF)

        if 1:
            #normalise variance
            data.normaliseY()
            dump_file = dump_file + 'norm'

    [data.N,data.D]    = data.Y.shape
    if 1:
        R1=test_diverse4(data,Nf=Nf)
        cPickle.dump(R1,open(os.path.join(out_base,dump_file+'.pickle'),'wb'))
        cPickle.dump(R1[0].results,open(os.path.join(out_base,dump_file+'.results.pickle'),'wb')) 
    if 0:
        #load dumped results and plot
        dump_file = 'smith_thesis3norm.pickle'
        R = cPickle.load(open(os.path.join(out_base,dump_file),'rb'))
        plotResults(R)
        PL.show()
