# VBFA fillin:
# identical to vbfa however allows to fill-in missing values
# the main difference is that noise levels are constructed to be NxD and not updated.
# this is a hack to accommodate the limitations of variaitonal inference

from numpy import *
import numpy.random as random
import numpy.linalg as linalg
import scipy.special as special
import sys
import cPickle
import logging as L
import scipy as S
import pdb
#use everything from BayesNet
from bayesnet import *
from expressionnet import AExpressionModule
from vbfa import *



class CNodeSf(AVGaussNode):
    """fill-in version of CNodeS"""
    def __init__(self,net,prec=1):
        AVGaussNode.__init__(self,dim=[net._N,net.components],cdim=1)

    def update(self,net=None):
        if(net==None):
            AVGaussNode.update(self)
            return
        W = net.W
        Eps=net.Eps
        #iterate through all individuals
        for n in range(net._N):
            #1. calculate the covariance matrix for this individual (because of fill-in this is individual-specific
            p = eye(net.components)+tensordot(Eps.E1[n,:],W.E2,[0,0])
            self.cov = linalg.inv(p)
            matrix   = dot( dot(self.cov,W.E1.T), diag(Eps.E1[n,:]) )
            self.E1[n,:]  = dot(net.Z.E1[n,:],matrix.T)
        AVGaussNode.update(self)



class CNodeWf(AVGaussNode):
    """fill-in version of CNodeW"""
    def __init__(self,net,E1,E2):
        self.E1 = E1
        self.E2 = E2


    def update(self,net=None):
        return



class CNodeEpsf(AGammaNode):
    """fill-in version of CNodeEps"""
    def __init__(self,net,Epsa,Epsb,mask):
        #intialise a structure of the size of mask
        self.a = zeros(mask.shape)
        self.b = zeros(mask.shape)
        self.a[:,:] = Epsa
        self.b[:,:] = Epsb
        #mask denotes non-missing values
        #make everything else missing by assigning a really high uncertainty
        #all that matters is that E1 = b/a becomes tiny,i.e. 0:-)
        self.b[~mask] = 0
        AGammaNode.update(self)
        pass
        
            


class CVBFA_fillin(CVBFA):
    '''CVBFA_fillin
     - hacked up VBFA to fill-in missing values...'''


    def getDefaultParameters(self):
        """return a hash with default parameter value for this BayesNet"""
        #
        dp = AExpressionModule.getDefaultParameters(self)
        dp['initType'] = 'pca'
        dp['nIterations'] = 20
        dp['schedule'] = ['S']
        dp['components'] = 5
        dp['priors']  = {}
        return dp


    def init(self,init_data,model_train,mask):
        """init_data: data to fill-in
           mask     : binary mask inidicating *observed* values
           model_train: VBFA type model to get the marginals fro Eps and W from
        """
        
        if not isinstance(init_data,AGaussNode):
            raise Exception("initialization is only possible from a GaussNode")
        self.Z = CNodeZ(node=init_data)
        self.dataNode = self.Z

        #adopt number of components
        self.components = model_train.components
        

        # set dimensionality of the data
        [self._N, self._D] = self.Z.E1.shape

        W_prior   = model_train.W
        Eps_prior = model_train.Eps

        #add the new nodes - to be replaced by XML init:
        self.nodes = {'S':CNodeSf(self),'W':CNodeWf(self,W_prior.E1,W_prior.E2),'Eps':CNodeEpsf(self,Eps_prior.a,Eps_prior.b,mask)} 
        for n in self.nodes.keys(): setattr(self,n,self.nodes[n])
        # no initialisation required!





class CMean_fillin():
    '''Mean model:  just model the mean of expression data and nothing else
     - hacked up VBFA to fill-in missing values...'''


    def init(self,init_data,Pi=None,model_train=None,mask=None):
        """init_data: data to fill-in
           mask     : binary mask inidicating *observed* values
           model_train: VBFA type model to get the marginals fro Eps and W from
        """
        if model_train is None:
            self.mean = init_data.E1.mean(axis=0)
            pass
        else:
            self.mean = model_train.mean
        
        pass

    def getName(self):
        """return a name summarising the  main parameters"""
        name = "Mean"
        return name

    def iterate(self):
        return

    def getPrediction(self):
        E1 = self.mean
        return CGauss(E1=E1)
        pass

